from importlib import import_module

__all__ = ("import_item",)


def import_item(name: str):
    """Import an item from a module"""

    match name.rsplit(".", 1):
        case [module_name, item_name]:
            module = import_module(module_name)
            return getattr(module, item_name)
        case _:
            raise ValueError("name must be in 'module.item' format")



default_settings = {
    "components": [],
    "middleware": [],
    "logging": {
        "setup": "selva.logging.setup.setup_logger",
        "root": "WARNING",
        "level": {},
        "enable": [],
        "disable": [],
    },
}



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable

from asgikit.requests import Request

__all__ = ("Middleware",)


class Middleware(ABC):
    @abstractmethod
    async def __call__(
        self,
        call: Callable[[Request], Awaitable],
        request: Request,
    ):
        raise NotImplementedError()



import inspect
from collections.abc import Callable
from typing import TypeVar

from selva.di.inject import Inject
from selva.di.service.model import InjectableType, ServiceInfo

__all__ = ("service", "DI_SERVICE_ATTRIBUTE")

DI_SERVICE_ATTRIBUTE = "__selva_di_service__"

T = TypeVar("T")


def _is_inject(value) -> bool:
    return isinstance(value, Inject) or value is Inject


def service(
    injectable: T = None, /, *, provides: type = None, name: str = None
) -> T | Callable[[T], T]:
    """Declare a class or function as a service

    For classes, a constructor will be generated to help create instances
    outside the dependency injection context
    """

    def inner(injectable: InjectableType) -> T:
        setattr(injectable, DI_SERVICE_ATTRIBUTE, ServiceInfo(provides, name))

        if inspect.isclass(injectable):
            annotations = set(inspect.get_annotations(injectable).keys())
            dependencies = [d for d, _ in inspect.getmembers(injectable, _is_inject)]

            if missing := [d for d in dependencies if d not in annotations]:
                # TODO: create exception
                raise TypeError(
                    f"Missing type annotation for dependencies of service '{injectable.__qualname__}': "
                    ", ".join(missing)
                )

            # save a reference to the original constructor
            original_init = getattr(injectable, "__init__", None)

            def init(self, *args, **kwargs):
                """Generated init method for service

                Positional and keywork arguments will be set to declared dependencies.
                Dependencies without an argument to set their value will be None.
                Remaining arguments will be ignored.
                """

                # call original constructor
                if original_init:
                    original_init(self)

                positional_params = [d for d in dependencies if d not in kwargs]

                # set positional argument values
                values = dict(zip(positional_params, args))

                # set keyword argument values
                values |= {k: v for k, v in kwargs.items() if k in dependencies}

                # the rest of the dependencies will be set to None
                param_keys = values.keys()
                values |= {d: None for d in dependencies if d not in param_keys}

                for k, v in values.items():
                    setattr(self, k, v)

            setattr(injectable, "__init__", init)

        return injectable

    return inner(injectable) if injectable else inner



import os
from collections import UserDict
from copy import deepcopy
from pathlib import Path
from typing import Any

import strictyaml
from loguru import logger

from selva.configuration.defaults import default_settings
from selva.configuration.environment import (
    parse_settings_from_env,
    replace_variables_recursive,
)

__all__ = ("Settings", "SettingsError", "get_settings")

SETTINGS_DIR_ENV = "SELVA_SETTINGS_DIR"
SETTINGS_FILE_ENV = "SELVA_SETTINGS_FILE"

DEFAULT_SETTINGS_DIR = str(Path("configuration"))
DEFAULT_SETTINGS_FILE = "settings.yaml"

SELVA_PROFILE = "SELVA_PROFILE"


class Settings(UserDict):
    def __init__(self, data: dict):
        for key, value in data.items():
            if isinstance(value, dict):
                data[key] = Settings(value)

        super().__init__(data)

    def __getattr__(self, item: str):
        try:
            return self.data[item]
        except KeyError:
            raise AttributeError(item)


class SettingsError(Exception):
    def __init__(self, path: Path):
        super().__init__(f"cannot load settings from {path}")
        self.path = path


def get_settings() -> Settings:
    # get default settings
    settings = deepcopy(default_settings)

    # merge with main settings file (settings.yaml)
    merge_recursive(settings, get_settings_for_profile())

    # merge with environment settings file (settings_$SELVA_ENV.yaml)
    if active_env := os.getenv(SELVA_PROFILE):
        merge_recursive(settings, get_settings_for_profile(active_env))

    # merge with environment variables (SELVA_*)
    from_env_vars = parse_settings_from_env(os.environ)
    merge_recursive(settings, from_env_vars)

    settings = replace_variables_recursive(settings, os.environ)
    return Settings(settings)


def get_settings_for_profile(env: str = None) -> dict[str, Any]:
    settings_file = os.getenv(SETTINGS_FILE_ENV, DEFAULT_SETTINGS_FILE)
    settings_dir_path = Path(os.getenv(SETTINGS_DIR_ENV, DEFAULT_SETTINGS_DIR))
    settings_file_path = settings_dir_path / settings_file

    if env is not None:
        settings_file_path = settings_file_path.with_stem(
            f"{settings_file_path.stem}_{env}"
        )

    settings_file_path = settings_file_path.absolute()

    try:
        settings_yaml = settings_file_path.read_text("utf-8")
        return strictyaml.load(settings_yaml).data
    except FileNotFoundError:
        logger.info("settings file not found: {}", settings_file_path)
        return {}
    except (KeyError, ValueError):
        raise
    except Exception as err:
        raise SettingsError(settings_file_path) from err


def merge_recursive(destination: dict, source: dict):
    for key in source:
        if key in destination and all(
            isinstance(arg[key], dict) for arg in (destination, source)
        ):
            merge_recursive(destination[key], source[key])
        else:
            destination[key] = deepcopy(source[key])
